/*
 * ProcessWatcher: プロセス監視クラス
 */

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;

public final class ProcessWatcher
{
	//
	// 内部型定義
	//

	// OutputState: 結果出力種別
	// 『Effective Java』「項目21 enum構文をクラスで置き換える」に従って実装
	public static final class OutputState
	{
		//
		// クラス属性
		//

		public static final OutputState OUT	= new OutputState("Out");
		public static final OutputState ERR	= new OutputState("Err");
		public static final OutputState WATCHER	= new OutputState("Watcher");

		//
		// 属性
		//
		private final String name;

		//
		// 構築子
		//
		private OutputState(
			String strName)	// 名前
		{
			 name = strName;
		}
	}

	// レポート受信インターフェース
	interface ReportReceiver
	{
		void receiveReport(int nId, OutputState state, String strOutput);
	}
	
	// ストリーム監視クラス
	private final class StreamWatcher implements Runnable
	{
		//
		// 属性
		//
		private final InputStream	watchStream;
		private final boolean	isError;

		//
		// 構築子
		//
		public StreamWatcher(
			InputStream stream,	// 監視対称ストリーム(入力)
			boolean bError)	// エラー出力監視フラグ
		{
			if (stream == null) {
				throw new NullPointerException();
			}
			
			watchStream = stream;
			isError = bError;
		}

		//
		// 再定義メソッド
		//
		
		// 実行する
		public void run()
		{
			boolean bSuccess = true;
			try {
				BufferedReader reader = new BufferedReader(new InputStreamReader(watchStream));
				while (watching) {
					String strOutput = reader.readLine();
					if (strOutput == null) {
						break;
					}

					reportReceiver.receiveReport(id,
												(isError ? OutputState.ERR : OutputState.OUT),
												strOutput);
				}
			}
			catch (IOException exp) {
				System.err.println(exp.getMessage());
				bSuccess = false;
			}
			finally {
				if (watching) {
					ProcessWatcher.this.setResult(bSuccess);
				}
			}
		}
	}

	//
	// クラス属性
	//
	static private volatile boolean	watching	= true;
	static private int	nextId;

	//
	// クラスメソッド
	//

	// 監視を停止する
	static public void stopWatch()
	{
		watching = false;
	}
		
	// IDを返す
	static private synchronized int getNextId()
	{
		return ++nextId;
	}
		
	//
	// 属性
	//

	private final Process	watchProcess;
	private final ReportReceiver	reportReceiver;
	private final int	id;
	private final Thread[] streamWatchers	= new Thread[2];
	private int	endStreamWatcherCount;

	//
	// 構築子
	//

	public ProcessWatcher(
		Process process,			// 監視対象プロセス(入力)
		String strName,				// プロセス名(入力)
		ReportReceiver receiver)	// 結果通知対象フレーム
	{
		if (process == null || strName == null || receiver == null) {
			throw new NullPointerException();
		}

		watchProcess = process;
		reportReceiver = receiver;

		id = getNextId();
		streamWatchers[0] = new Thread(new StreamWatcher(watchProcess.getInputStream(), false));
		streamWatchers[1] = new Thread(new StreamWatcher(watchProcess.getErrorStream(), true));
		for (int nIndex = 0; nIndex < streamWatchers.length; nIndex++) {
			streamWatchers[nIndex].start();
		}

		StringBuffer sb = new StringBuffer();
		sb.append("Start '");
		sb.append(strName);
		sb.append("' as id:" + id);

		reportReceiver.receiveReport(id, OutputState.WATCHER, sb.toString());
	}

	//
	// 実装
	//

	// 結果を設定する
	private synchronized void setResult(
		boolean bSuccess)	// 成否フラグ
	{
		if (bSuccess) {
			endStreamWatcherCount++;
		}

		if (endStreamWatcherCount == streamWatchers.length) {
			StringBuffer sb = new StringBuffer();
			sb.append("Exit(Code=");
			try {
				sb.append(watchProcess.waitFor());
			}
			catch (InterruptedException exp) {
				System.err.println(exp.getMessage());
				sb.append("(");
				sb.append(exp.getMessage());
				sb.append(")");
			}
			finally {
				sb.append(")");
			}

			reportReceiver.receiveReport(id,
										OutputState.WATCHER,
										sb.toString());
		}
	}
}
