/*
 * MainFrame: メインフレーム
 * (BridgeパターンのRefinedAbstractionクラス) 
 * (Template MethodパターンのConcreteClassクラス) 
 */

import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Frame;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import ui.ActionManager;
import environment.EnvInfo;

import view.ImplementOperation;

public final class MainFrame extends ui.MainFrameBase
	implements ClipboardOwner, ProcessWatcher.ReportReceiver 
{
	//
	// 内部定義型
	//

	// ViewState: 表示種別
	// 『Effective Java』「項目21 enum構文をクラスで置き換える」に従って実装
	public static final class ViewState
	{
		//
		// クラス属性
		//

		public static final ViewState LIST	= new ViewState("List");
		public static final ViewState EDIT	= new ViewState("Edit");
		public static final ViewState RESULT	= new ViewState("Result");
		public static final ViewState JRE	= new ViewState("JRE");
		public static final ViewState INFO	= new ViewState("Info");

		//
		// 属性
		//
		private final String name;

		//
		// コンストラクタ
		//
		private ViewState(
			String strName)	// 名前
		{
			 name = strName;
		}
	}

	// 実行アクション
	public final class ExecuteAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			execute();
		}
	}

	// 登録アクション
	public final class RegistAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			regist();
		}
	}

	// 編集アクション
	public final class EditAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			edit();
		}
	}

	// 更新アクション
	public final class UpdateAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			update();
		}
	}

	// 移動(上に)アクション
	public final class MoveUpAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			swap(false);
		}
	}

	// 移動(下に)アクション
	public final class MoveDownAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			swap(true);
		}
	}

	// 削除アクション
	public final class DeleteAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			delete();
		}
	}

	// コピーアクション
	public final class CopyAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			copy();
		}
	}

	// スクリプト作成アクション
	public final class CreateScriptAction implements ActionListener
	{
		// アクションを実行する
		public void actionPerformed(
			ActionEvent e)	// アクションイベント
		{
			createScript();
		}
	}

	// ステータスバー表示状態設定アクション
	public final class SetStatusBarVisibleAction implements ItemListener
	{
		// アクションを実行する
		public void itemStateChanged(
			ItemEvent e)	// イベント
		{
			final boolean bVisible = (e.getStateChange() == ItemEvent.SELECTED);
			setStatusBarVisible(bVisible);
		}
	}

	// 出力報告 
	private final class OutputReporter implements Runnable
	{
		private String output;
		
		public OutputReporter(
			String strOutput)
		{
			if (strOutput == null || strOutput.length() == 0) {
				throw new IllegalArgumentException();
			}
			output = strOutput;
		}
		
		// 実行する
		public void run()
		{
			reportResult(output);
		}
	}
	
	//
	// クラス属性
	//
	public static final String RESKEY_WIN_SELECTFILE	= "SelectExecuteFileDialog";
	public static final String RESKEY_WIN_ADDTCLASSPATH	= "AddClassPathDialog";
	public static final String RESKEY_WIN_SCRIPTFILEPATH	= "CreateScriptFileDialog";

	public static final String PROPKEY_STATUSBAR = "ViewStatusBar";

	private static final char CHAR_RUNINFOSEPARATOR = '\t';
	private static final MainFrame instance = new MainFrame();

	//
	// クラスメソッド(公開)
	//

	// 唯一のインスタンスを返す
	public static MainFrame getInstance()
	{
		return instance;
	}

	// システムプロパティを返す
	public static String[] getSystemPropertiesInfo()
	{
		Properties properties = System.getProperties();

		String[] astrProperties = new String[properties.size()];
		int nIndex = 0;
		StringBuffer sb = new StringBuffer();
		Enumeration enum = properties.propertyNames();
		while (enum.hasMoreElements()) {
			String strPropertyName = (String)enum.nextElement(); 

			sb.append(strPropertyName);
			sb.append(':');
			sb.append(EnvInfo.NEWLINE);
			String strValue =System.getProperty(strPropertyName);
			if (strPropertyName.equals("line.separator")) {
				// 改行コードはそのまま出すとなにがなんだかわからないので十六進表記に変換する
				int nCharIndex = 0;
				while (nCharIndex < strValue.length()) {
					String strHexValue = "0" + Integer.toHexString(strValue.charAt(nCharIndex)); 
					sb.append("0x");
					sb.append(strHexValue.substring(strHexValue.length() - 2));
					
					nCharIndex++;
					if (nCharIndex < strValue.length()) {
						sb.append('+');
					}
				}
			}
			else if (strValue == null || strValue.length() == 0) {
				sb.append("(no value)");
			}
			else {
				sb.append(strValue);
			}
			sb.append(EnvInfo.NEWLINE);

			astrProperties[nIndex] = sb.toString();
			
			sb.setLength(0); 
			nIndex++;
		}

		Arrays.sort(astrProperties);		
		return astrProperties;
	}

	//
	// クラスメソッド(実装)
	//

	// 設定ファイルの保存パスを返す
	public static String makeConfFilePath()
	{
		return EnvInfo.getInstance().getUserHomeDir()
				+ File.separatorChar
				+ "sbkick.list";
	}

	// プロセスの出力報告時の前置文字列を作成する 
	private static String makePrefix(
		int nId,		// ID
		char charType)	// 出力種別
	{
		StringBuffer sb = new StringBuffer();
		sb.append(nId);
		sb.append(':');
		if (charType != '\0') {
			sb.append(charType);
		}
		sb.append('>');

		return sb.toString();
	}
	
	//
	// 属性
	//
	public final ActionListener executeAction	= new ExecuteAction();
	public final ActionListener registAction	= new RegistAction();
	public final ActionListener editAction	= new EditAction();
	public final ActionListener updateAction	= new UpdateAction();
	public final ActionListener moveUpAction	= new MoveUpAction();
	public final ActionListener moveDownAction	= new MoveDownAction();
	public final ActionListener deleteAction	= new DeleteAction();
	public final ActionListener copyAction	= new CopyAction();
	public final ActionListener createScriptAction	= new CreateScriptAction();
	public final ItemListener setStatusBarVisibleAction	= new SetStatusBarVisibleAction();

	private ActionManager actionManager;
	private final List runInfoList	= new ArrayList();
	private int reportedLines;

	//
	// コンストラクタ
	//
	private MainFrame()
	{
		super(true);
	}

	//
	// 操作
	//

	// ステータスバーの表示状態を設定する
	void setStatusBarVisible(
		boolean bVisible)	// 表示フラグ
	{
		ImplementOperation op = getOperation();
		if (op != null) {
			op.setStatusBarVisible(bVisible);
		}
	}

	// ステータスバーの表示状態を設定する
	public void setStatusBarVisibleCheckStatus(
		boolean bCheck)	// チェックフラグ
	{
		actionManager.setCheckStatus(setStatusBarVisibleAction, bCheck);
	}
	
	//
	// 実装
	//

	// コントローラを初期化する
	private void initController(
		Frame frame)	// 実装(フレーム)
	{
		ControllerDefinition definition = new ControllerDefinition();
		definition.buildMenu(this, frame, actionManager);
	}

	// 実装を初期化する
	private void initImplementor(
		Frame frame)	// 実装(フレーム)
	{
		// フォントを設定する
		frame.setFont(EnvInfo.getInstance().getBaseFont(frame));

		initController(frame);

		ImplementOperation imp = (ImplementOperation)frame;
		imp.initComponent();
	}

	// フレームをImplementOperationインターフェースとして返す
	private ImplementOperation getOperation()
	{
		return (ImplementOperation)getImplementor();
	}

	// 選択されている実行情報を返す
	private RunProgramInfo getSelectedRunInfo()
	{
		int nSelectedIndex = getOperation().getSelectedRunInfoIndex();
		return (0 <= nSelectedIndex
				? (RunProgramInfo)runInfoList.get(nSelectedIndex)
				: null);
	}

	// 操作対象となる実行情報を返す
	private RunProgramInfo getOperationTargetRunInfo()
	{
		final ImplementOperation op = getOperation();
		final ViewState state = op.getCurrentViewState();
		if (state != ViewState.LIST && state != ViewState.EDIT) {
			throw new IllegalStateException();
		}

		return (state == ViewState.LIST
				? getSelectedRunInfo()
				: op.getEditingRunProgramInfo());
	}

	// 実行する
	private void execute()
	{
		final ImplementOperation op = getOperation();
		final ViewState viewState = op.getCurrentViewState();
		if (viewState != ViewState.EDIT && viewState != ViewState.LIST) {
			beep();
			return;
		}

		// 実行情報を取得する
		RunProgramInfo runInfo;
		if (viewState == ViewState.EDIT) {
			runInfo = op.getEditingRunProgramInfo();
			if (runInfo == null) {
				op.showMessage("実行に必要な情報が不足しています。");
				return;
			}
		}
		else {
			runInfo = getSelectedRunInfo();
			if (runInfo == null) {
				op.showMessage("実行情報が選択されていません。");
				return;
			}
		}

		RunProgramInfo.State state = runInfo.getState();
		if (state != RunProgramInfo.State.NORMAL) {
			String strMsg;
			if (state == RunProgramInfo.State.NO_PROGRAM) {
				strMsg = "実行プログラムが指定されていません。";
			}
			else if (state == RunProgramInfo.State.PROGRAM_NOT_EXIST) {
				strMsg = "実行用に指定されたファイルは存在しません。";
			}
			else if (state == RunProgramInfo.State.CANNOT_EXECUTE) {
				strMsg = "実行できないJavaファイルです。";
			}
			else if (state == RunProgramInfo.State.PATH_NOT_EXIST) {
				strMsg = "クラスパスに存在しないパスが指定されています。";
			}
			else {
				strMsg = state.name;
			}
	
			op.showMessage(strMsg);
			return;
		}
	
		setWaitCursor();
		try {
			String strRunName = runInfo.getName();
			if (strRunName == null || strRunName.length() == 0) {
				strRunName = runInfo.getCommandLine();
			}

			new ProcessWatcher(runInfo.execute(),
								strRunName,
								(ProcessWatcher.ReportReceiver)this);
			op.selectView(ViewState.RESULT);

			restoreCursor();
		}
		catch (IOException e) {
			restoreCursor();
			showExceptionMessage(e);
		}
		catch (Exception e) {
			restoreCursor();
			showExceptionMessage(e);
		}
	}

	// 結果を報告する
	private void reportResult(
		String strResult)	// 出力行
	{
		if (strResult == null) {
			throw new IllegalArgumentException();
		}

		boolean bDeleteResult = (1000 <= reportedLines);
		getOperation().reportResult(strResult, bDeleteResult);
		if (bDeleteResult == false) {
			reportedLines++;
		}

		System.out.println(strResult);
	}
	
	// 登録する
	private void regist()
	{
		final ImplementOperation op = getOperation();
		if (op.getCurrentViewState() != ViewState.EDIT) {
			beep();
			return;
		}

		try {
			RunProgramInfo runInfo = op.getEditingRunProgramInfo();
			if (runInfo == null || runInfo.getName().length() == 0) {
				op.showMessage("登録に必要な情報が不足しています。");
				return;
			}

			runInfoList.add(runInfo);

			// 本当は末尾への追加だけでいい。
			op.setRunInfoViewList(runInfoList);

			op.selectView(ViewState.LIST);
		}
		catch (Exception e) {
			showExceptionMessage(e);
		}
	}

	// 編集する
	private void edit()
	{
		final ImplementOperation op = getOperation();
		if (op.getCurrentViewState() != ViewState.LIST) {
			beep();
			return;
		}

		try {
			RunProgramInfo runInfo = getSelectedRunInfo();
			if (runInfo == null) {
				beep();
				return;
			}

			// 編集画面に情報を渡す
			op.setEditRunProgramInfo(runInfo);
			op.selectView(ViewState.EDIT);
		}
		catch (Exception e) {
			showExceptionMessage(e);
		}
	}

	// 更新する
	private void update()
	{
		final ImplementOperation op = getOperation();
		if (op.getCurrentViewState() != ViewState.EDIT) {
			beep();
			return;
		}
	
		try {
			int nSelectedIndex = op.getSelectedRunInfoIndex();
			if (nSelectedIndex < 0) {
				beep();
				return;
			}
	
			// 更新を確認する
			StringBuffer sb = new StringBuffer();
			sb.append("実行情報「");
			sb.append(getSelectedRunInfo().getName());
			sb.append("」を更新します。よろしいですか?");
			if (op.showConfirmMessage(sb.toString()) == false) {
				return;
			}
	
			RunProgramInfo runInfo = op.getEditingRunProgramInfo();
			if (runInfo == null || runInfo.getName().length() == 0) {
				op.showMessage("更新に必要な情報が不足しています。");
				return;
			}

			runInfoList.set(nSelectedIndex, runInfo);

			// 本当は更新した部分だけ変更すればいい。
			op.setRunInfoViewList(runInfoList);
			op.setListSelectedIndex(nSelectedIndex);

			op.selectView(ViewState.LIST);
		}
		catch (Exception e) {
			showExceptionMessage(e);
		}
	}

	// ひとつ前またはうしろに移動する
	private void swap(
		boolean bDown)	// 下方向移動フラグ
	{
		final ImplementOperation op = getOperation();
		if (op.getCurrentViewState() != ViewState.LIST) {
			beep();
			return;
		}

		try {
			int nSelectedIndex = op.getSelectedRunInfoIndex();
			if (nSelectedIndex < 0) {
				beep();
				return;
			}
		
			int nSwapIndex = nSelectedIndex + (bDown ? 1 : -1);
			if (nSwapIndex < 0 || runInfoList.size() <= nSwapIndex) {
				beep();
				return;
			}

			RunProgramInfo infoSelected = (RunProgramInfo)runInfoList.get(nSelectedIndex);
			RunProgramInfo infoChange = (RunProgramInfo)runInfoList.get(nSwapIndex);
			runInfoList.set(nSelectedIndex, infoChange);
			runInfoList.set(nSwapIndex, infoSelected);

			// 本当は変更した部分だけ反映させればいい
			op.setRunInfoViewList(runInfoList);
			op.setListSelectedIndex(nSwapIndex);
		}
		catch (Exception e) {
			showExceptionMessage(e);
		}
	}
	
	// 削除する
	private void delete()
	{
		final ImplementOperation op = getOperation();
		if (op.getCurrentViewState() != ViewState.LIST) {
			beep();
			return;
		}

		try {
			int nSelectedIndex = op.getSelectedRunInfoIndex();
			if (nSelectedIndex < 0) {
				beep();
				return;
			}

			// 削除を確認する
			StringBuffer sb = new StringBuffer();
			sb.append("実行情報「");
			sb.append(getSelectedRunInfo().getName());
			sb.append("」を削除します。よろしいですか?");

			if (op.showConfirmMessage(sb.toString()) == false) {
				return;
			}

			runInfoList.remove(nSelectedIndex);

			// 本当は削除した部分だけ変更すればいい
			op.setRunInfoViewList(runInfoList);
		}
		catch (Exception e) {
			showExceptionMessage(e);
		}
	}

	// テキストをクリップボードにコピーする
	private void copy()
	{
		final ImplementOperation op = getOperation();
		try {
			String strTextToCopy = null;
			final ViewState state = op.getCurrentViewState();
			if (state == ViewState.LIST || state == ViewState.EDIT) {
				RunProgramInfo runInfo = getOperationTargetRunInfo();
				if (runInfo != null && runInfo.getState() == RunProgramInfo.State.NORMAL) {
					strTextToCopy = runInfo.getCommandLine();
				}
			}
			else {
				strTextToCopy = op.getTextToCopyToClipboard();
			}

			boolean bSuccess = false;
			if (strTextToCopy != null) {
				Clipboard clipboaard = getSystemClipboard();
				if (clipboaard != null) { 
					clipboaard.setContents(new StringSelection(strTextToCopy), this);
					bSuccess = true;
				}
			}

			if (bSuccess != false) {
				String strCopyTarget = "";
				if (state == ViewState.LIST) {
					strCopyTarget = "コマンドライン";
				}
				else if (state == ViewState.EDIT) {
					strCopyTarget = "コマンドライン";
				}
				else if (state == ViewState.RESULT) {
					strCopyTarget = "結果";
				}
				else if (state == ViewState.JRE) {
					strCopyTarget = "クラス情報";
				}
				else if (state == ViewState.INFO) {
					strCopyTarget = "情報";
				}

				op.setStatusText(strCopyTarget + "をコピーしました。");
			}
			else {
				beep();
			}
		}
		catch (Exception e) {
			showExceptionMessage(e);
		}
	}

	// スクリプトを作成する
	private void createScript()
	{
		final ImplementOperation op = getOperation();
		final ViewState state = op.getCurrentViewState();
		if (state != ViewState.LIST && state != ViewState.EDIT) {
			beep();
			return;
		}

		RunProgramInfo runInfo = getOperationTargetRunInfo();
		if (runInfo == null || runInfo.getState() != RunProgramInfo.State.NORMAL) {
			beep();
			return;
		}

		try {
			EnvInfo env =  EnvInfo.getInstance();

			// デフォルトのファイル名を作成する
			String strDefFileName = runInfo.getName();
			String strDefFileExt = env.getDefaultScriptExt();
			if (0 < strDefFileExt.length()) {
				strDefFileName += "." + strDefFileExt;
			}

			// スクリプトファイルの保存パスを取得する
			String strScriptFileName = op.getCreateScriptFilePath(env.getUserHomeDir(), strDefFileName);
			if (0 < strScriptFileName.length()) {
				// ファイルに保存する
				env.createCommandScriptFile(runInfo.getCommandLine(),
											strScriptFileName);
											
				op.setStatusText("スクリプトファイル" + strScriptFileName + "を作成しました。");
			}
		}
		catch (Exception e) {
			showExceptionMessage(e);
		}
	}

	//
	// 再定義メソッド(MainFrameBase)
	//

	// デフォルトのサイズを返す
	protected Dimension getDefaultSize()
	{
		Frame frame = getImplementor();
		if (frame == null) {
			throw new IllegalStateException();
		}

		frame.pack();
		return frame.getSize();
	}

	// 追加状態を復元する
	protected void restoreAdditionalState()
	{
		try {
			getOperation().setAdditionalState(getStateProperties());
		}
		catch (Exception exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}
	}

	// 追加状態を設定する
	protected void setAdditionalStatePropreties()
	{
		Map mapAdditionalState = getOperation().getAdditionalState();
		Set keySet = mapAdditionalState.keySet();
		Iterator it = keySet.iterator();
		while (it.hasNext()) {
			String strKey = (String)it.next();
			setStatePropertyValue(strKey, (String)mapAdditionalState.get(strKey));
		}
	}

	// アプリケーション名を返す
	protected String getAppName(
		boolean bShort)	//  短縮名フラグ
	{
		return (bShort ? "sbkick" : "StompBox/Kick");
	}

	// 実装を作成して返す
	// 注記: J2MEに存在しないクラスのロードを避けるためにリフレクションで作成する
	protected Frame createImplementor(
		ArgumentInfo argument)	// 引数情報
	{
		Frame frame = null;

		try {
			final boolean bAwt = isToolkitAwt(argument);
			if (bAwt == false) {
				initSwing(argument);
			//	frame = new view.swing.MainFrameImp();
				frame = (Frame)Class.forName("view.swing.MainFrameImp").newInstance();
			}
			else {
				frame = new view.awt.MainFrameImp();
			//	frame = (Frame)Class.forName("view.awt.MainFrameImp").newInstance();
			}
			actionManager = ActionManager.createInstance(bAwt);

			initImplementor(frame);
		}
		catch (ClassNotFoundException exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}
		catch (IllegalAccessException exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}
		catch (InstantiationException exp) {
			// Do nothing.
			outputExceptionInfo(exp);
		}

		return frame;		
	}

	// 初期化する
	protected boolean initialize(
		ArgumentInfo argument)	// 引数情報
	{
		boolean bSuccess = true;

		setTitle(getAppName(false) + " Ver.1.4");

		// アイコンを設定する
		setIcon();

		// 設定を読みこむ
		File confFile = new File(makeConfFilePath());
		if (confFile.exists()) {
			BufferedReader reader = null;
			try {
				final FileInputStream fis = new FileInputStream(confFile);
				reader = new BufferedReader(new InputStreamReader(fis, EnvInfo.getInstance().getDefaultEncoding()));
				String strLine;
				while ((strLine = reader.readLine()) != null) {
					runInfoList.add(RunProgramInfo.createInstanceFromLine(strLine,
																			CHAR_RUNINFOSEPARATOR));
				}
			}
			catch (IOException e) {
				showExceptionMessage(e);
				bSuccess = false;
			}
			finally {
				if (reader != null) {
					try {
						reader.close();
					}
					catch (IOException e) {
						// Do nothing.
						outputExceptionInfo(e);
					}
				}
			}
		}

		// 実装に設定をセットする
		if (bSuccess != false) {
			final ImplementOperation op = getOperation();
			op.setRunInfoViewList(runInfoList);
			op.selectView(runInfoList.size() == 0 ? ViewState.EDIT : ViewState.LIST);
		}

		return bSuccess;
	}

	// 終了可能かを返す 
	protected boolean canExit()
	{
		// 常に終了可とする
		return true;
	}

	// 終了準備処理を行う
	protected void prepareExit()
	{
		// 監視を終了する
		ProcessWatcher.stopWatch();

		// 設定内容を記録する
		BufferedWriter writer = null;
		try {
			String strConfFilePath = makeConfFilePath();
			if (0 < runInfoList.size()) {
				FileOutputStream fos = new FileOutputStream(strConfFilePath);
				OutputStreamWriter osw = new OutputStreamWriter(fos, EnvInfo.getInstance().getDefaultEncoding());
				writer = new BufferedWriter(osw);
				Iterator it = runInfoList.iterator();
				while (it.hasNext()) {
					writer.write((String)((RunProgramInfo)it.next()).toLine(CHAR_RUNINFOSEPARATOR));
					writer.newLine();
				}
			}
			else {
				File fileConf = new File(strConfFilePath);
				if (fileConf.exists()) { 
					fileConf.delete();
				}
			}
		}
		catch (IOException e) {
			showExceptionMessage(e);
		}
		finally {
			if (writer != null) {
				try {
					writer.flush();
					writer.close();
				}
				catch (IOException e) {
					// Do nothing.
					outputExceptionInfo(e);
				}
			}
		}
	}

	//
	// 再定義メソッド(ClipboardOwner)
	//

	// クリップボードに転送したデータの所有権喪失を受信する
	public void lostOwnership(
		Clipboard clipboard,	// クリップボード
		Transferable contents)	// 転送した内容
	{
		// Do nothing.
	}

	//
	// 再定義メソッド(ProcessWatcher.ReportReceiver)
	//

	// プロセス監視結果を受信する
	public void receiveReport(
		int nId,			// プロセスのID
		ProcessWatcher.OutputState state,
		String strOutput)	// 出力文字列
	{
		char cPrefix = '-';
		if (state == ProcessWatcher.OutputState.OUT) {
			cPrefix = 'O';
		}
		else if (state == ProcessWatcher.OutputState.ERR) {
			cPrefix = 'E';
		}
		Runnable reporter = new OutputReporter(makePrefix(nId, cPrefix) + strOutput);
		EventQueue.invokeLater(reporter);
	}
}
