/*
 * RunProgramInfo: 実行情報
 */

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import environment.EnvInfo;
import io.JavaFileUtilities;

public final class RunProgramInfo
{
	//
	// 内部定義型
	//

	// State: 状態
	// 『Effective Java』「項目21 enum構文をクラスで置き換える」に従って実装
	public static final class State
	{
		//
		// クラス属性
		//

		public static final State NORMAL	= new State("Normal");
		public static final State NO_PROGRAM	= new State("No program");
		public static final State PROGRAM_NOT_EXIST	= new State("Program not exist");
		public static final State CANNOT_EXECUTE	= new State("Cannot execute");
		public static final State PATH_NOT_EXIST	= new State("Path not exist");

		//
		// 属性
		//
		public final String name;

		//
		// 構築子
		//
		private State(
			String strName)	// 名前
		{
			 name = strName;
		}
	}
	
	//
	// 属性
	//

	private final String name;
	private final String program;
	private final String argument;
	private final Set classPath;

	//
	// クラスメソッド
	//

	// パラメータ行からインスタンスを作成する
	static RunProgramInfo createInstanceFromLine(
		String strLine,	// 作成元となる行
		char cSplit)	// 区切り文字
	{
		if (strLine == null) {
			throw new NullPointerException();
		}

		int n1stSplitPos = strLine.indexOf(cSplit);
		if (n1stSplitPos < 0) {
			throw new IllegalArgumentException();
		}
		int n2ndSplitPos = strLine.indexOf(cSplit, n1stSplitPos + 1);
		if (n2ndSplitPos < 0) {
			throw new IllegalArgumentException();
		}
		int n3rdSplitPos = strLine.indexOf(cSplit, n2ndSplitPos + 1);

		String strName = strLine.substring(0, n1stSplitPos);
		String strProgram = strLine.substring(n1stSplitPos + 1, n2ndSplitPos);
		String strArgument;

		Set setClassPath = new TreeSet();
		if (n2ndSplitPos < n3rdSplitPos) {
			strArgument = strLine.substring(n2ndSplitPos + 1, n3rdSplitPos);

			String strClassPath = strLine.substring(n3rdSplitPos + 1);
			int nSplitPos = 0;
			while (0 < (nSplitPos = strClassPath.indexOf(File.pathSeparator))) {
				setClassPath.add(strClassPath.substring(0, nSplitPos));
				strClassPath = strClassPath.substring(nSplitPos + 1);
			}
			setClassPath.add(strClassPath);
		}
		else {
			strArgument = strLine.substring(n2ndSplitPos + 1);
		}

		return new RunProgramInfo(strName, strProgram, strArgument, setClassPath);
	}

	//
	// 構築子
	//
	public RunProgramInfo(
		String strName,				// 名前 
		String strProgram,			// 実行ファイル
		String strArgument,			// 引数
		Collection colClassPath)	// クラスパスの集合
	{
		if (strName == null || strProgram == null || strArgument == null) {
			throw new NullPointerException();
		}

		name = strName;
		program = strProgram;
		argument = strArgument;

		if (EnvInfo.getInstance().isPathCharCaseInsensitive()) {
			classPath = new TreeSet(String.CASE_INSENSITIVE_ORDER);
			classPath.addAll(colClassPath);
		}
		else {
			classPath = new TreeSet(colClassPath);
		}
	}

	//
	// 状態
	//

	// 名前を返す
	public String getName()
	{
		return name;
	}

	// プログラムを返す
	public String getProgram()
	{
		return program;
	}

	// Javaプログラムかを返す
	public boolean isJavaProgram()
	{
		return JavaFileUtilities.isExecutableJavaFile(new File(program));
	}

	// 引数を返す
	public String getArgument()
	{
		return argument;
	}

	// クラスパスの集合を返す
	public Set getClassPath()
	{
		return new TreeSet(classPath);
	}

	// コマンドライン文字列を作成して返す
	public String getCommandLine()
		throws IOException
	{
		if (getState() != State.NORMAL) {
			throw new IllegalStateException();
		}
		
		String strCmdLine;
		if (isJavaProgram()) {
			strCmdLine = EnvInfo.getInstance().makeJavaCmdLine(program, classPath, argument, name);
		}
		else {
			StringBuffer sb = new StringBuffer();
			sb.append(program);
			if (argument != null && 0 < argument.length()) {
				sb.append(' ');
				sb.append(argument);
			}
			strCmdLine = sb.toString();
		}
		return strCmdLine;
	}

	// 状態を返す
	public State getState()
	{
		if (program.length() == 0) {
			// プログラムが指定されていない
			return State.NO_PROGRAM;
		}

		File file = new File(program);
		if (file.exists() == false || file.isFile() == false) {
			// 指定プログラムは存在しない
			return State.PROGRAM_NOT_EXIST;
		}

		JavaFileUtilities.FileType type = JavaFileUtilities.getFileType(file);
		if (type == JavaFileUtilities.FileType.JAR
		|| type == JavaFileUtilities.FileType.CLASS) {
			if (JavaFileUtilities.isExecutableJavaFile(file) == false) {
				// 実行できないJavaファイル
				return State.CANNOT_EXECUTE;
			}

			Iterator it = classPath.iterator();
			while (it.hasNext()) {
				if (new File((String)it.next()).exists() == false) {
					// クラスパスは存在しない
					return State.PATH_NOT_EXIST;
				}
			}
		}

		return State.NORMAL;
	}

	// 行文字列を作成して返す
	public String toLine(
		char cConnect)	// 接続文字
	{
		StringBuffer sb = new StringBuffer();
		sb.append(name);
		sb.append(cConnect);
		sb.append(program);
		sb.append(cConnect);
		sb.append(argument);

		if (0 < classPath.size()) {
			sb.append(cConnect);
			sb.append(EnvInfo.makeJavaClassPathLine(classPath));
		}

		return sb.toString();
	}

	//
	// 操作
	//

	// 実行する	
	public Process execute()
		throws IOException
	{
		if (getState() != State.NORMAL) {
			throw new IllegalStateException();
		}

		String strCmdLine = getCommandLine();
		System.err.println("Execute: " + strCmdLine);
		return Runtime.getRuntime().exec(strCmdLine);
	}
}
