/*
 * MainFrameImp: Swing版メインウィンドウ
 * 注記: BridgeパターンのImplementorクラス
 */
package view.swing;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.text.BadLocationException;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import environment.EnvInfo;
import io.JavaFileUtilities;
import ui.swing.LookAndFeelUtilities;

import ControllerDefinition;
import MainFrame;
import RunProgramInfo;
import view.ImplementOperation;
import view.LabelManager;

public final class MainFrameImp extends JFrame
	implements ImplementOperation
{
	//
	// 内部型定義
	//

	// Java実行ファイルフィルタ
	private static class ExecutableFileFilter extends FileFilter
	{
		//
		// 状態
		//

		// 受け入れ可能ファイルかを返す
		public boolean accept(
			File file)	// 判定対象ファイル 
		{
			return (file.exists()
					&& (file.isDirectory()
						|| JavaFileUtilities.isExecutableJavaFile(file)));
		}

		// 記述を返す 		
		public String getDescription()
		{
			return "Java実行ファイル(*.class;*.jar)";
		}
	}
	
	// スクリプトファイルフィルタ
	private static class ScriptFileFilter extends FileFilter
	{
		//
		// 状態
		//

		// 受け入れ可能ファイルかを返す
		public boolean accept(
			File file)	// 判定対象ファイル 
		{
			return (file.isDirectory() == false);
		}

		// 記述を返す 		
		public String getDescription()
		{
			return "スクリプトファイル(*." + EnvInfo.getInstance().getDefaultScriptExt() + ")";
		}
	}

	//
	// クラス属性
	//

	// テキストエリアのマージン
	private static final Insets MARGIN_TEXTAREA = new Insets(0, 4, 0, 4);

	//
	// クラスメソッド(実装)
	//

	// テキストエリアを作成する
	private static JTextArea createTextArea(
		Font font)	// フォント
	{
		JTextArea textArea = new JTextArea();
		textArea.setFont(font);
		textArea.setMargin(MARGIN_TEXTAREA);
		return textArea;
	}

	//
	// 属性
	//

	private List lafMenuItems;

	private JTabbedPane tab;

	private JList listRunInfoName;
	private JList listClassPath;

	private JTextField textFile;
	private JTextArea textArgument;
	private JTextField textName;
	private JButton buttonRegist;
	private JButton buttonUpdate;

	private JTextArea	textResult;

	private JTextArea	textClassInfo;

	private JTextArea	textSysInfo;

	private JLabel labelStatus;

	//
	// コンストラクタ
	//
	public MainFrameImp()
	{
		super();
	}

	//
	// 実装
	//

	// 「実行」ボタンを作成する
	private JButton createExecButton(
		Properties properties)	// 文字列リソースプロパティ
	{
		JButton buttonExec = new JButton(properties.getProperty(LabelManager.LABEL_BTN_EXEC));
		buttonExec.addActionListener(MainFrame.getInstance().executeAction);

		return buttonExec;
	}
	
	// 「一覧」パネルの内容を初期化する
	private void initSelectPanel(
		Properties properties)	// 文字列リソースプロパティ
	{
		listRunInfoName = new JList();
		listRunInfoName.setFont(EnvInfo.getInstance().getBaseFont(this));
		listRunInfoName.setModel(new DefaultListModel());

		JScrollPane scrollPane = new JScrollPane();
		scrollPane.getViewport().setView(listRunInfoName);

		JPanel panelSelect = new JPanel(new BorderLayout());
		panelSelect.add(scrollPane, BorderLayout.CENTER);

		JPanel panelButton = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;
		c.gridx = 0;
		c.gridy = 0;

		JButton buttonUp = new JButton(properties.getProperty(LabelManager.LABEL_BTN_UP));
		buttonUp.addActionListener(MainFrame.getInstance().moveUpAction);
		panelButton.add(buttonUp, c);

		JButton buttonDown = new JButton(properties.getProperty(LabelManager.LABEL_BTN_DOWN));
		buttonDown.addActionListener(MainFrame.getInstance().moveDownAction);
		c.gridx++;
		panelButton.add(buttonDown, c);

		c.gridy++;
		c.gridx = 0;
		c.gridwidth = 2;
		panelButton.add(createExecButton(properties), c);

		panelSelect.add(panelButton, BorderLayout.SOUTH);

		tab.addTab(properties.getProperty(LabelManager.LABEL_PANE_SELECT), panelSelect);
	}
	
	// 「入力」パネルの内容を初期化する
	private void initInputPanel(
		Properties properties)	// 文字列リソースプロパティ
	{
		Font fontBase = EnvInfo.getInstance().getBaseFont(this);

		textFile = new JTextField();
		textFile.setColumns(40);
		textFile.setFont(fontBase);

		JButton refFileBtn = new JButton(properties.getProperty(LabelManager.LABEL_BTN_REFER));
		refFileBtn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				selectExecuteFile();
			}
		});

		listClassPath = new JList();
		listClassPath.setFont(fontBase);
		listClassPath.setVisibleRowCount(3);
		listClassPath.setModel(new DefaultListModel());

		JScrollPane scrollPane = new JScrollPane();
		scrollPane.setPreferredSize(textFile.getPreferredSize());
		scrollPane.getViewport().setView(listClassPath);

		JButton refCPBtn = new JButton(properties.getProperty(LabelManager.LABEL_BTN_ADD));
		refCPBtn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				addClassPath();
			}
		});

		JButton delCPBtn = new JButton(properties.getProperty(LabelManager.LABEL_BTN_DELETE));
		delCPBtn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				removeClassPath();
			}
		});

		textArgument = createTextArea(fontBase);
		textArgument.setRows(3);

		textName = new JTextField();
		textName.setColumns(textFile.getColumns());
		textName.setFont(fontBase);

		buttonRegist = new JButton(properties.getProperty(LabelManager.LABEL_BTN_REGIST));
		buttonRegist.addActionListener(MainFrame.getInstance().registAction);

		buttonUpdate = new JButton(properties.getProperty(LabelManager.LABEL_BTN_UPDATE));
		buttonUpdate.addActionListener(MainFrame.getInstance().updateAction);

		Component[] components = new Component[] {
			new JLabel(properties.getProperty(LabelManager.LABEL_FILE)),
			textFile,
			refFileBtn,
			new JLabel(properties.getProperty(LabelManager.LABEL_CP)),
			scrollPane,
			refCPBtn,
			delCPBtn,
			new JLabel(properties.getProperty(LabelManager.LABEL_ARG)),
			new JScrollPane(textArgument),
			new JLabel(properties.getProperty(LabelManager.LABEL_NAME)),
			textName,
			buttonRegist,
			buttonUpdate,
			createExecButton(properties),
		};

		JPanel panelInput = new JPanel();
		view.awt.MainFrameImp.setInputTabPaneLayout(components, panelInput);
		tab.addTab(properties.getProperty(LabelManager.LABEL_PANE_INPUT), panelInput);
	}

	// 「結果」パネルの内容を初期化する
	private void initResultPanel(
		Properties properties)	// 文字列リソースプロパティ
	{
		JPanel panelResult = new JPanel();
		panelResult.setLayout(new GridLayout(1, 1));
		textResult = createTextArea(EnvInfo.getInstance().getBaseFont(this));
		textResult.setEditable(false);
		textResult.setLineWrap(true);
		panelResult.add(new JScrollPane(textResult));
		tab.addTab(properties.getProperty(LabelManager.LABEL_PANE_RESULT), panelResult);
	}
	
	// 「JRE」パネルの内容を初期化する
	private void initJrePanel(
		Properties properties)		// 文字列リソースプロパティ
	{
		final JPanel panelJre = new JPanel();

		final JTextField field = new JTextField();
		final JButton button = new JButton(properties.getProperty(LabelManager.LABEL_BTN_SHOW));
		button.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				final Cursor saveCursor = getCursor();
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				textClassInfo.setText(view.ClassInfoSerializer.get(field.getText()));
				textClassInfo.setCaretPosition(0);
				setCursor(saveCursor);
			}
		});
		
		textClassInfo = createTextArea(EnvInfo.getInstance().getBaseFont(this));
		textClassInfo.setEditable(false);

		final Component[] components = new Component[] {
			new JLabel(properties.getProperty(LabelManager.LABEL_CLASS)),
			field,
			button,
			new JScrollPane(textClassInfo),
		};
		view.awt.MainFrameImp.setJreTabPaneLayout(components, panelJre);
		tab.addTab(properties.getProperty(LabelManager.LABEL_PANE_JRE), panelJre);
	}
	
	// 「情報」パネルの内容を初期化する 
	private void initSysInfoPanel(
		Properties properties)	// 文字列リソースプロパティ
	{
		JPanel panelSysInfo = new JPanel();
		panelSysInfo.setLayout(new GridLayout(1, 1));
		textSysInfo = createTextArea(EnvInfo.getInstance().getBaseFont(this));
		textSysInfo.setEditable(false);
		textSysInfo.setLineWrap(true);
		panelSysInfo.add(new JScrollPane(textSysInfo));
		tab.addTab(properties.getProperty(LabelManager.LABEL_PANE_INFO), panelSysInfo);
	}

	// タブ選択処理を実行する
	private void selectTab()
	{
		if (getCurrentViewState() == MainFrame.ViewState.INFO
		 && textSysInfo.getText().length() == 0) {
			final Cursor saveCursor = getCursor();
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

			textSysInfo.setText(view.awt.MainFrameImp.getSystemPropertyInfoText());
			textSysInfo.setCaretPosition(0);

			setCursor(saveCursor);
		}
	}

	// ファイル参照処理を行う
	private void selectExecuteFile()
	{
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogTitle(MainFrame.getFrameTitle(MainFrame.RESKEY_WIN_SELECTFILE));
		fileChooser.setSelectedFile(new File(textFile.getText()));
		fileChooser.setFileFilter(new ExecutableFileFilter());
		if (fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			textFile.setText(fileChooser.getSelectedFile().getPath());
		}
	}

	// クラスパス設定処理を行う
	private void addClassPath()
	{
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogTitle(MainFrame.getFrameTitle(MainFrame.RESKEY_WIN_ADDTCLASSPATH));
		fileChooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
		if (fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			File fileSelected = fileChooser.getSelectedFile();
			DefaultListModel model = (DefaultListModel)listClassPath.getModel();
			if (model.contains(fileSelected.getPath()) == false) {
				model.addElement(fileSelected.getPath());

				int nLastIndex = model.size() - 1;
				listClassPath.setSelectedIndex(nLastIndex);
				listClassPath.ensureIndexIsVisible(nLastIndex);
			}
		}
	}

	// クラスパス削除処理を行う
	private void removeClassPath()
	{
		final int nSelectedIndex = listClassPath.getSelectedIndex(); 
		if (0 <= nSelectedIndex) {
			DefaultListModel model = (DefaultListModel)listClassPath.getModel();
			model.remove(nSelectedIndex);
		}
		else {
			getToolkit().beep();
		}
	}

	//
	// 再定義メソッド(MainFrameBaseImplementor)
	//

	// コンポーネントを初期化する
	public void initComponent()
	{
		// メニューに見た目変更用メニューを追加する
		List list = LookAndFeelUtilities.addSelectMenuItem(this);
		if (1 < list.size()) {
			lafMenuItems = list;

			JMenu menuView = getJMenuBar().getMenu(ControllerDefinition.MENU_VIEW); 
			menuView.addSeparator();
			Iterator it = lafMenuItems.iterator();
			while (it.hasNext()) {
				JRadioButtonMenuItem item = (JRadioButtonMenuItem)it.next();
				menuView.add(item);
			}
		}

		// レイアウトマネージャを設定する
		Container container = getContentPane();
		container.setLayout(new BorderLayout(4, 4));

		final Properties properties = LabelManager.getComponentLabelProperties();
		tab = new JTabbedPane();
		tab.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				selectTab();
			}
		});
		container.add(tab, BorderLayout.CENTER);

		// 「選択」タブの内容を作成する
		initSelectPanel(properties);
	
		// 「指定」タブの内容を作成する
		initInputPanel(properties);

		// 「結果」タブの内容を作成する
		initResultPanel(properties);

		// 「JRE」タブの内容を作成する
		initJrePanel(properties);

		// 「情報」タブの内容を作成する
		initSysInfoPanel(properties);

		labelStatus = new JLabel(properties.getProperty(LabelManager.LABEL_STATUS_READY));
		container.add(labelStatus, BorderLayout.SOUTH);
	}

	// メッセージダイアログを表示する
	public void showMessage(
		String strMessage)	// メッセージ文字列(入力)
	{
		getToolkit().beep();
		JOptionPane.showMessageDialog(this, strMessage, getTitle(), JOptionPane.INFORMATION_MESSAGE);
	}

	//
	// 再定義メソッド(ImplementOperation)
	//

	// 確認メッセージを表示する
	public boolean showConfirmMessage(
		String strMessage)	// メッセージ文字列
	{
		getToolkit().beep();
		return (JOptionPane.showConfirmDialog(this,
												strMessage,
												getTitle(),
												JOptionPane.OK_CANCEL_OPTION) == JOptionPane.OK_OPTION); 	
	}
	
	// 現在の表示状態を返す 
	public MainFrame.ViewState getCurrentViewState()
	{
		MainFrame.ViewState state;
		switch (tab.getSelectedIndex()) {
			case 0:
				state = MainFrame.ViewState.LIST; 
				break;
			case 1:
				state = MainFrame.ViewState.EDIT; 
				break;
			case 2:
				state = MainFrame.ViewState.RESULT; 
				break;
			case 3:
				state = MainFrame.ViewState.JRE; 
				break;
			case 4:
				state = MainFrame.ViewState.INFO; 
				break;
			default:
				throw new IllegalArgumentException();
		}
		return state;
	}

	// 編集中の実行情報を返す 
	public RunProgramInfo getEditingRunProgramInfo()
	{
		RunProgramInfo info = null;
		if (0 < textFile.getText().length()) {
			DefaultListModel model = (DefaultListModel)listClassPath.getModel();
			info = new RunProgramInfo(textName.getText(),
										textFile.getText(),
										textArgument.getText().replace('\n', ' '),
										Arrays.asList(model.toArray()));
		}
		return info; 
	}

	// 編集中の実行情報を返す 
	public int getSelectedRunInfoIndex()
	{
		return listRunInfoName.getSelectedIndex();
	}

	// クリップボードにコピーするテキストを返す
	public String getTextToCopyToClipboard()
	{
		String strTextToCopy = null;

		JTextArea textArea = null;
		MainFrame.ViewState state = getCurrentViewState();
		if (state == MainFrame.ViewState.RESULT) {
			textArea = textResult;
		}
		else if (state == MainFrame.ViewState.JRE) {
			textArea = textClassInfo;
		}
		else if (state == MainFrame.ViewState.INFO) {
			textArea = textSysInfo;
		}
		
		if (textArea != null) {
			if (textArea.getSelectionStart() < textArea.getSelectionEnd()) {
				strTextToCopy = textArea.getSelectedText();
			}
			else {
				strTextToCopy = textArea.getText();
			}
		}

		return strTextToCopy;
	}

	// 作成スクリプトのファイルパスを返す
	public String getCreateScriptFilePath(
		String strDefaultPath,	// デフォルトのパス
		String strDefaultName)	// デフォルトのファイル名
	{
		if (strDefaultPath == null) {
			throw new NullPointerException();
		}
		if (strDefaultName == null) {
			throw new NullPointerException();
		}

		String strScriptFilePath = "";

		JFileChooser fileChooser = new JFileChooser(strDefaultPath);
		fileChooser.setDialogTitle(MainFrame.getFrameTitle(MainFrame.RESKEY_WIN_SCRIPTFILEPATH));
		fileChooser.setSelectedFile(new File(strDefaultName));
		fileChooser.setFileFilter(new ScriptFileFilter());
		if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			strScriptFilePath = fileChooser.getSelectedFile().getPath();
		}

		return strScriptFilePath;
	}
	
	// 追加の状態を返す
	public Map getAdditionalState()
	{
		Map mapState = new HashMap();
		if (lafMenuItems != null) {
			mapState.putAll(LookAndFeelUtilities.getMenuItemState(lafMenuItems));
		}
		mapState.put(MainFrame.PROPKEY_STATUSBAR, Boolean.toString(labelStatus.isVisible()));
		return mapState;
	}

	// 追加の状態を設定する
	public void setAdditionalState(
		Map properties)	// 状態プロパティ
	{
		if (properties == null) {
			throw new IllegalArgumentException();
		}

		if (lafMenuItems != null) {
			LookAndFeelUtilities.setMenuItemState(properties, lafMenuItems);
		}
		if (properties.containsKey(MainFrame.PROPKEY_STATUSBAR)) {
			final boolean bVisible = Boolean.valueOf((String)properties.get(MainFrame.PROPKEY_STATUSBAR)).booleanValue();
			MainFrame.getInstance().setStatusBarVisibleCheckStatus(bVisible);
		}
	}
	
	// 表示を切り替える
	public void selectView(
		MainFrame.ViewState viewState)	// 表示状態
	{
		final int nSelectedIndex;
		if (viewState == MainFrame.ViewState.LIST) {
			nSelectedIndex = 0;
		}
		else if (viewState == MainFrame.ViewState.EDIT) {
			nSelectedIndex = 1;
		}
		else if (viewState == MainFrame.ViewState.RESULT) {
			nSelectedIndex = 2;
		}
		else {
			throw new IllegalArgumentException();
		}

		if (tab.getSelectedIndex() != nSelectedIndex) {
			tab.setSelectedIndex(nSelectedIndex);
		}
	}

	// 実行情報の一覧を設定する	
	public void setRunInfoViewList(
		List runInfoList)	// 実行情報の一覧
	{
		// AbstractListModelインターフェースを実装したクラスにrunInfoListを
		// 結びつけるという実現方法もいつか試してみたいがいまはこれがせいいっぱい
		if (runInfoList == null) {
			throw new NullPointerException();
		}

		DefaultListModel listModel = (DefaultListModel)listRunInfoName.getModel();

		// リストをクリアする
		listModel.clear();

		// リストに名前を設定する
		Iterator it = runInfoList.iterator();
		while (it.hasNext()) {
			RunProgramInfo info = (RunProgramInfo)it.next();
			listModel.addElement(info.getName());
		}
	}

	// 編集実行情報を設定する
	public void setListSelectedIndex(
		int nIndex)	// 選択インデックス
	{
		listRunInfoName.setSelectedIndex(nIndex);
		listRunInfoName.ensureIndexIsVisible(nIndex);
	}

	// 編集実行情報を設定する
	public void setEditRunProgramInfo(
		RunProgramInfo runInfo)	// 実行情報
	{
		if (runInfo == null) {
			throw new IllegalArgumentException();
		}

		textName.setText(runInfo.getName());
		textFile.setText(runInfo.getProgram());
		textArgument.setText(runInfo.getArgument());
		
		DefaultListModel model = (DefaultListModel)listClassPath.getModel();
		model.clear();
		
		Set setClassPath = runInfo.getClassPath();
		Iterator it = setClassPath.iterator();
		while (it.hasNext()) {
			model.addElement(it.next());
		}
	}

	// 結果を設定する
	public void reportResult(
		String strResult,		// 結果文字列
		boolean bDeleteResult)	// 結果削除フラグ
	{
		if (strResult == null) {
			throw new IllegalArgumentException();
		}

		try {
			if (bDeleteResult) {
				textResult.replaceRange("", 0, textResult.getLineEndOffset(0));
			}
		}
		catch (BadLocationException exp) {
			// Do nothing.
			exp.printStackTrace();
		}

		textResult.append(strResult + EnvInfo.NEWLINE);
		setStatusText(strResult);
	}

	// ステータスバーにテキストを設定する
	public void setStatusText(
		String strText)	// 設定するテキスト
	{
		labelStatus.setText(strText);
	}

	// ステータスバーの表示状態を設定する
	public boolean setStatusBarVisible(
		boolean bVisible)	// 表示フラグ
	{
		labelStatus.setVisible(bVisible);
		labelStatus.invalidate();
		validate();

		return bVisible;
	}
}
